module("L_Concord", package.seeall)

local nixio = require("nixio")

-------------------------------------------------------------------------------
-- Flags
-------------------------------------------------------------------------------

local DEBUG_MODE = false
local CREATE_OK = false
local EXTRA_FEATURES = false
local QUICK_ARM = true


-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------

local VERSION = "1.0"

local SEND_DELAY = 2 -- The delay after which to send the next message in the queue.

-- Services
local CONCORD_SID   = "urn:micasaverde-com:serviceId:ConcordAlarmPanel1"
local PARTITION_SID = "urn:micasaverde-com:serviceId:AlarmPartition2"
local SES_SID       = "urn:micasaverde-com:serviceId:SecuritySensor1"

-- Message types
local TASK_BUSY       =  1
local TASK_BUSY_PERM  = -1
local TASK_ERROR      =  2
local TASK_ERROR_PERM = -2
local TASK_SUCCESS    =  4

-- Different characters
local CHAR_CR  = string.char(0xD)
local CHAR_LF  = string.char(0xA)
local CHAR_ACK = string.char(0x6)
local CHAR_NAK = string.char(0x15)

-- Zone statuses
local ZONE_TRIPPED  =  1
local ZONE_FAULTED  =  2
local ZONE_TROUBLE  =  8
local ZONE_BYPASSED = 16

-- Panel features
local FEATURE_CHIME         =  1
local FEATURE_ENERGY_SAVER  =  2
local FEATURE_NO_DELAY      =  4
local FEATURE_LATCHKEY      =  8
local FEATURE_SILENT_ARMING = 16
local FEATURE_QUICK_ARM     = 32

local COMMANDS = {
	-- Arming modes
	["Disarmed"] = "01",
	["Stay"] = "02",
	["StayInstant"] = "02",
	["StaySilent"] = "02",
	["Armed"] = "03",
	["ArmedInstant"] = "03",
	["ArmedSilent"] = "03",
	-- Panic alarms
	["Police"] = "4C",
	["Auxiliary"] = "4D",
	["Fire"] = "4E",
	-- Chime mode
	["Chime"] = "0701",
	-- Bypass / arm zone
	["Bypass"] = "0B"
}

local PANEL_TYPES = {
	["14"] = "Concord",
	["0B"] = "Concord Express",
	["1E"] = "Concord Express 4",
	["0E"] = "Euro"
}

local TOKENS = {
	["00"] = "0",
	["01"] = "1",
	["02"] = "2",
	["03"] = "3",
	["04"] = "4",
	["05"] = "5",
	["06"] = "6",
	["07"] = "7",
	["08"] = "8",
	["09"] = "9",
	["0C"] = "#",
	["0D"] = ":",
	["0E"] = "/",
	["0F"] = "?",
	["10"] = ".",
	["11"] = "A",
	["12"] = "B",
	["13"] = "C",
	["14"] = "D",
	["15"] = "E",
	["16"] = "F",
	["17"] = "G",
	["18"] = "H",
	["19"] = "I",
	["1A"] = "J",
	["1B"] = "K",
	["1C"] = "L",
	["1D"] = "M",
	["1E"] = "N",
	["1F"] = "O",
	["20"] = "P",
	["21"] = "Q",
	["22"] = "R",
	["23"] = "S",
	["24"] = "T",
	["25"] = "U",
	["26"] = "V",
	["27"] = "W",
	["28"] = "X",
	["29"] = "Y",
	["2A"] = "Z",
	["2B"] = " ",
	["2C"] = "'",
	["2D"] = "-",
	["2E"] = "_",
	["2F"] = "*",
	["30"] = "AC POWER",
	["31"] = "ACCESS",
	["32"] = "ACCOUNT",
	["33"] = "ALARM",
	["34"] = "ALL",
	["35"] = "ARM",
	["36"] = "ARMING",
	["37"] = "AREA",
	["38"] = "ATTIC",
	["39"] = "AUTO",
	["3A"] = "AUXILIARY",
	["3B"] = "AWAY",
	["3C"] = "BACK",
	["3D"] = "BATTERY",
	["3E"] = "BEDROOM",
	["3F"] = "BEEPS",
	["40"] = "BOTTOM",
	["41"] = "BREEZEWAY",
	["42"] = "BASEMENT",
	["43"] = "BATHROOM",
	["44"] = "BUS",
	["45"] = "BYPASS",
	["46"] = "BYPASSED",
	["47"] = "CABINET",
	["48"] = "CANCELED",
	["49"] = "CARPET",
	["4A"] = "CHIME",
	["4B"] = "CLOSET",
	["4C"] = "CLOSING",
	["4D"] = "CODE",
	["4E"] = "CONTROL",
	["4F"] = "CPU",
	["50"] = "DEGREES",
	["51"] = "DEN",
	["52"] = "DESK",
	["53"] = "DELAY",
	["54"] = "DELETE",
	["55"] = "DINING",
	["56"] = "DIRECT",
	["57"] = "DOOR",
	["58"] = "DOWN",
	["59"] = "DOWNLOAD",
	["5A"] = "DOWNSTAIRS",
	["5B"] = "DRAWER",
	["5C"] = "DISPLAY",
	["5D"] = "DURESS",
	["5E"] = "EAST",
	["5F"] = "ENERGY SAVER",
	["60"] = "ENTER",
	["61"] = "ENTRY",
	["62"] = "ERROR",
	["63"] = "EXIT",
	["64"] = "FAIL",
	["65"] = "FAILURE",
	["66"] = "FAMILY",
	["67"] = "FEATURES",
	["68"] = "FIRE",
	["69"] = "FIRST",
	["6A"] = "FLOOR",
	["6B"] = "FORCE",
	["6C"] = "FORMAT",
	["6D"] = "FREEZE",
	["6E"] = "FRONT",
	["6F"] = "FURNACE",
	["70"] = "GARAGE",
	["71"] = "GALLERY",
	["72"] = "GOODBYE",
	["73"] = "GROUP",
	["74"] = "HALL",
	["75"] = "HEAT",
	["76"] = "HELLO",
	["77"] = "HELP",
	["78"] = "HIGH",
	["79"] = "HOURLY",
	["7A"] = "HOUSE",
	["7B"] = "IMMEDIATE",
	["7C"] = "IN SERVICE",
	["7D"] = "INTERIOR",
	["7E"] = "INTRUSION",
	["7F"] = "INVALID",
	["80"] = "IS",
	["81"] = "KEY",
	["82"] = "KITCHEN",
	["83"] = "LAUNDRY",
	["84"] = "LEARN",
	["85"] = "LEFT",
	["86"] = "LIBRARY",
	["87"] = "LEVEL",
	["88"] = "LIGHT",
	["89"] = "LIGHTS",
	["8A"] = "LIVING",
	["8B"] = "LOW",
	["8C"] = "MAIN",
	["8D"] = "MASTER",
	["8E"] = "MEDICAL",
	["8F"] = "MEMORY",
	["90"] = "MIN",
	["91"] = "MODE",
	["92"] = "MOTION",
	["93"] = "NIGHT",
	["94"] = "NORTH",
	["95"] = "NOT",
	["96"] = "NUMBER",
	["97"] = "OFF",
	["98"] = "OFFICE",
	["99"] = "OK",
	["9A"] = "ON",
	["9B"] = "OPEN",
	["9C"] = "OPENING",
	["9D"] = "PANIC",
	["9E"] = "PARTITION",
	["9F"] = "PATIO",
	["A0"] = "PHONE",
	["A1"] = "POLICE",
	["A2"] = "POOL",
	["A3"] = "PORCH",
	["A4"] = "PRESS",
	["A5"] = "QUIET",
	["A6"] = "QUICK",
	["A7"] = "RECEIVER",
	["A8"] = "REAR",
	["A9"] = "REPORT",
	["AA"] = "REMOTE",
	["AB"] = "RESTORE",
	["AC"] = "RIGHT",
	["AD"] = "ROOM",
	["AE"] = "SCHEDULE",
	["AF"] = "SCRIPT",
	["B0"] = "SEC",
	["B1"] = "SECOND",
	["B2"] = "SET",
	["B3"] = "SENSOR",
	["B4"] = "SHOCK",
	["B5"] = "SIDE",
	["B6"] = "SIREN",
	["B7"] = "SLIDING",
	["B8"] = "SMOKE",
	["B9"] = "Sn",
	["BA"] = "SOUND",
	["BB"] = "SOUTH",
	["BC"] = "SPECIAL",
	["BD"] = "STAIRS",
	["BE"] = "START",
	["BF"] = "STATUS",
	["C0"] = "STAY",
	["C1"] = "STOP",
	["C2"] = "SUPERVISORY",
	["C3"] = "SYSTEM",
	["C4"] = "TAMPER",
	["C5"] = "TEMPERATURE",
	["C6"] = "TEMPORARY",
	["C7"] = "TEST",
	["C8"] = "TIME",
	["C9"] = "TIMEOUT",
	["CA"] = "TOUCHPAD",
	["CB"] = "TRIP",
	["CC"] = "TROUBLE",
	["CD"] = "UNBYPASS",
	["CE"] = "UNIT",
	["CF"] = "UP",
	["D0"] = "VERIFY",
	["D1"] = "VIOLATION",
	["D2"] = "WARNING",
	["D3"] = "WEST",
	["D4"] = "WINDOW",
	["D5"] = "MENU",
	["D6"] = "RETURN",
	["D7"] = "POUND",
	["D8"] = "HOME"
}

local ALARM_TYPES = {
	["00"] = "Unspecified",
	["01"] = "Fire",
	["02"] = "Fire Panic",
	["03"] = "Police",
	["04"] = "Police Panic",
	["05"] = "Medical",
	["06"] = "Medical Panic",
	["07"] = "Auxiliary",
	["08"] = "Auxiliary Panic",
	["09"] = "Tamper",
	["10"] = "No Activity",
	["11"] = "Suspicion",
	["12"] = "Not used",
	["13"] = "Low Temperature",
	["14"] = "High Temperature",
	["15"] = "Keystroke Violation (Touchpad Tamper)",
	["16"] = "Duress",
	["17"] = "Exit Fault",
	["18"] = "Explosive Gas",
	["19"] = "Carbon Monoxide",
	["20"] = "Environmental",
	["21"] = "Latchkey",
	["22"] = "Equipment Tamper",
	["23"] = "Holdup",
	["24"] = "Sprinkler",
	["25"] = "Heat",
	["26"] = "Siren Tamper",
	["27"] = "Smoke",
	["28"] = "Repeater Tamper",
	["29"] = "Fire Pump Activated",
	["30"] = "Fire Pump Failure",
	["31"] = "Fire Gate Valve",
	["32"] = "Low CO2 Pressure",
	["33"] = "Low Liquid Pressure",
	["34"] = "Low Liquid Level",
	["35"] = "Entry/Exit",
	["36"] = "Perimeter",
	["37"] = "Interior",
	["38"] = "Near (Two Trip, Concord only)",
	["39"] = "Water Alarm"
}


-------------------------------------------------------------------------------
-- Globals
-------------------------------------------------------------------------------

local lug_device = nil

local g_taskHandle = -1
local g_lastTask = 0 -- Timestamp of the last status message.

local g_partitions = {}
local g_zones = {}

local messageQueue = {}
local sendAttempt = 0


-------------------------------------------------------------------------------
-- Generic utilities
-------------------------------------------------------------------------------

local log      = luup.log
local getinfo  = debug.getinfo -- Get the 'getinfo' function from the 'debug' module before overwriting the 'debug' keyword.
local checkBit = nixio.bit.check

local function debug (text)
	if (DEBUG_MODE == true) then
		log(text)
	end
end


local function task (text, mode)

	local messageType
	if (mode == TASK_ERROR or mode == TASK_ERROR_PERM) then
		messageType = "ERROR"
	else
		messageType = "STATUS"
	end
	debug("(ConcordPlugin::task/"..(getinfo(2).name or "N/A")..") "..messageType..": "..text)

	g_lastTask = os.time()

	if (mode == TASK_ERROR_PERM) then
		luup.task(text, TASK_ERROR, "Concord Plugin", g_taskHandle)
	elseif (mode == TASK_BUSY_PERM) then
		luup.task(text, TASK_BUSY, "Concord Plugin", g_taskHandle)
	else
		luup.task(text, mode, "Concord Plugin", g_taskHandle)
		luup.call_delay("clearTask", 30, tostring(g_lastTask)) -- Clear the previous message, since it's transient.
	end
end


function clearTask (taskTime)
	if (taskTime ~= nil and tostring(g_lastTask) ~= taskTime) then
		return
	end
	luup.task("Clearing...", TASK_SUCCESS, "Concord Plugin", g_taskHandle)
end


local function padLeft (s, length, c)
	s = tostring(s)
	length = length or 2
	c = c or "0"
	return c:rep(length - #s) .. s
end


function table.contains (t, item)
	for k, v in pairs(t) do
		if (v == item) then
			return true, k
		end
	end

	return false
end

-- Set "class".
local set = {
	add = function (s, item)
	          if (not table.contains(s, item)) then
	              s[#s + 1] = item
	          end
	      end,
	remove = function (s, item)
	             local item, pos = table.contains(t, item)
	             if (item) then
	                 table.remove(t, pos)
	             end
	         end
}


-------------------------------------------------------------------------------
-- Panel specific utilities
-------------------------------------------------------------------------------

local function getPartitionNumber (deviceId)
	for partNum, attr in pairs(g_partitions) do
		if (deviceId == attr.id) then
			return partNum
		end
	end
end


local function getZoneNumber (deviceId)
	for zoneNum, attr in pairs(g_zones) do
		if (deviceId == attr.id) then
			return zoneNum
		end
	end
end


local function checksum (s)
	--debug("(ConcordPlugin::checksum) s = " .. (s or "nil"))
	local total = 0
	for i = 1, #s, 2 do
		total = total + tonumber(s:sub(i, i+1), 16)
	end

	return total % 256
end


local function addToQueue(s)
	local length = padLeft(#s / 2 + 1)
	s = CHAR_LF .. length .. s .. string.format("%02X", checksum(length .. s))
	table.insert(messageQueue, s)
	if (#messageQueue == 1) then -- This is the only message in the queue. Start the 'send' chain.
		sendMessage()
	end
end


function sendMessage()
	local s
	if (#messageQueue > 0) then
		s = messageQueue[1]
	else
		return
	end

	sendAttempt = sendAttempt + 1
	if (sendAttempt == 3) then
		-- Send CR/LF to reset the automation module's message queue.
		if (not luup.io.write(CHAR_CR .. s)) then
			task("Failed to send message", TASK_ERROR_PERM)
			return
		end
		-- Remove the message from the queue, because there's a chance that the message is wrong.
		sendAttempt = 0
		table.remove(messageQueue, 1)
	elseif (not luup.io.write(s)) then
		-- Send the message normally.
		task("Failed to send message", TASK_ERROR_PERM)
		return
	end

	-- Send the next message in the queue.
	if (#messageQueue > 0) then
		luup.call_delay("sendMessage", SEND_DELAY)
	end
end


local function sendChar (c)
	if (not luup.io.write(c)) then
		task("Failed to send message", TASK_ERROR_PERM)
	end
end


local function decodeName (s)
	if (s == nil or s == "") then
		return "N/A"
	end

	s = s:gsub("(%x%x)", TOKENS)

	-- Trim leading and trailing whitespace.
	return s:match("^%s*(.-)%s*$")
end


-- Expands a number, e.g. 21 -> 0201
local function expand (nr)
	if (nr == nil) then
		return ""
	end

	local expNr = ""
	for i = 1, #nr do
		local d = nr:sub(i, i)
		expNr = expNr .. padLeft(d)
	end

	return expNr
end


-------------------------------------------------------------------------------
-- Message handler functions
-------------------------------------------------------------------------------

local function updateZone (zoneNum, zoneStatus)
	local zone = g_zones[zoneNum]

	if (zoneStatus ~= nil) then
		zone.tripped  = checkBit(zoneStatus, ZONE_TRIPPED)  and "1" or "0"
		zone.faulted  = checkBit(zoneStatus, ZONE_FAULTED)  and "1" or "0"
		zone.trouble  = checkBit(zoneStatus, ZONE_TROUBLE)  and "1" or "0"
		zone.bypassed = checkBit(zoneStatus, ZONE_BYPASSED) and "1" or "0"
	end

	if (zone.id ~= nil) then
		luup.variable_set(SES_SID, "Tripped", zone.tripped, zone.id)
		if (zone.tripped == "1") then
			luup.variable_set(SES_SID, "LastTrip", os.time(), zone.id)
		end
		luup.variable_set(CONCORD_SID, "Bypassed", zone.bypassed, zone.id)
		luup.variable_set(CONCORD_SID, "Faulted",  zone.faulted,  zone.id)
		luup.variable_set(CONCORD_SID, "Trouble",  zone.trouble,  zone.id)
	end
end


local function updatePartition (partNum, armingLevel)
	local partition = g_partitions[partNum]

	if (armingLevel ~= nil) then
		if (armingLevel == "03") then     -- Armed Away
			partition.armMode = "Armed"
			partition.detailedArmMode = "Armed"
		elseif (armingLevel == "02") then -- Armed Stay
			partition.armMode = "Armed"
			partition.detailedArmMode = "Stay"
		else                              -- Disarmed
			partition.armMode = "Disarmed"
			partition.detailedArmMode = "Disarmed"
		end
	end

	if (partition.id ~= nil) then
		luup.variable_set(PARTITION_SID, "ArmMode", partition.armMode, partition.id)
		luup.variable_set(PARTITION_SID, "DetailedArmMode", partition.detailedArmMode, partition.id)
	end
end


local function setupDevices()

	local ptr = luup.chdev.start(lug_device)

	-- Create zones.
	for zoneNum, zoneAttr in pairs(g_zones) do
		debug("(ConcordPlugin::MESSAGE_HANDLERS) Create/update zone #".. zoneNum ..".")
		local parameters = SES_SID ..",Armed=1\n".. SES_SID ..",Tripped=".. zoneAttr.tripped
		zoneNum = padLeft( zoneNum, 2)
		luup.chdev.append(lug_device, ptr, "concord_zone_"..zoneNum, "Zone #".. zoneNum .." - "..zoneAttr.name, "urn:schemas-micasaverde-com:device:MotionSensor:1", "D_MotionSensor1.xml", "", parameters, false)
	end

	-- Create partitions.
	for partNum, partAttr in pairs(g_partitions) do
		debug("(ConcordPlugin::MESSAGE_HANDLERS) Create/update partition #"..partNum..".")
		local parameters = PARTITION_SID ..",ArmMode=".. partAttr.armMode .."\n".. PARTITION_SID ..",DetailedArmMode=" ..partAttr.armMode
		luup.chdev.append(lug_device, ptr, "concord_partition_"..partNum, "Concord Partition #"..partNum, "urn:schemas-micasaverde-com:device:ConcordAlarmPartition:1", "D_ConcordPartition.xml", "", parameters, false)
	end

	luup.chdev.sync(lug_device, ptr)

	CREATE_OK = false

	-- Get the zones and partitions device ID.
	for id, attr in pairs(luup.devices) do
		if (attr.device_num_parent == lug_device) then
			local deviceType, num = attr.id:match("^concord_(%l+)_(%d+)$")
			if (deviceType == "zone") then
				num = tonumber(num, 10)
				g_zones[num].id = id
			elseif (deviceType == "partition") then
				num = tonumber(num, 10)
				g_partitions[num].id = id
			else
				log("(ConcordPlugin::MESSAGE_HANDLERS) Found bastart child, device #"..id..": "..attr.description..".")
			end
		end
	end

	-- Update partitions.
	for partNum in pairs(g_partitions) do
		updatePartition(partNum)
	end

	-- Update zones.
	for zoneNum in pairs(g_zones) do
		updateZone(zoneNum)
	end
end


local MESSAGE_HANDLERS = {
	["01"] = function (data)
				debug("(ConcordPlugin::MESSAGE_HANDLERS) 01 = Panel Type")
				local panelType = data:match("^(%d%d)")
				log("(ConcordPlugin::MESSAGE_HANDLERS) Panel type = " .. PANEL_TYPES[panelType])
				return true
			end,
	["03"] = function (data)
				debug("(ConcordPlugin::MESSAGE_HANDLERS) 03 = Zone Data")

				-- [PN] [AN] [GN] [ZNh] [ZNl] [ZT] [ZS] [text]
				local partNum, zoneNumHigh, zoneNumLow, zoneStatus, name = data:match("^(%x%x)%x%x%x%x(%x%x)(%x%x)%x%x(%x%x)(.*)$")
				partNum = tonumber(partNum, 16)
				local zoneNum = tonumber(zoneNumHigh .. zoneNumLow, 16)
				zoneStatus = tonumber(zoneStatus, 16)
				name = decodeName(name)
				debug("(ConcordPlugin::MESSAGE_HANDLERS) partNum = "..partNum..", zoneNum = "..zoneNum..", zoneStatus = "..zoneStatus..", name = '"..name.."'")

				-- Add entries in the Zones and Partitions tables.
				g_zones[zoneNum] = {
					name = name,
					partNum = partNum
				}
				updateZone(zoneNum, zoneStatus)

				CREATE_OK = false -- Not OK to create devices. We're missing partition data.
				return true
			end,
	["04"] = function (data)
				debug("(ConcordPlugin::MESSAGE_HANDLERS) 04 = Partition Data")

				-- [PN] [AN] [AL] [text]
				local partNum, armingLevel = data:match("^(%x%x)%x%x(%x%x)")
				partNum = tonumber(partNum, 16)

				if (g_partitions[partNum] ~= nil) then
					updatePartition(partNum, armingLevel)
					debug("(ConcordPlugin::MESSAGE_HANDLERS) Partition #" .. partNum .. " is enabled, armingLevel = " .. armingLevel .. ".")
				end

				CREATE_OK = true -- OK to create devices. We have all the data we need.
				return true
			end,
	["08"] = function (data)
				debug("(ConcordPlugin::MESSAGE_HANDLERS) 08 = Equipment List Complete")

				if (CREATE_OK) then -- Create partition and zone devices.
					debug("(ConcordPlugin::MESSAGE_HANDLERS) OK to create devices.")

					setupDevices()

					if (EXTRA_FEATURES) then
						addToQueue("20")
					end
				else
					debug("(ConcordPlugin::MESSAGE_HANDLERS) Not OK to create devices. Missing partition data.")
					-- Request partition data.
					addToQueue("0204")
				end

				clearTask()
				return true
			end,
	["21"] = function (data)
				debug("(ConcordPlugin::MESSAGE_HANDLERS) 21 = Zone Status")

				-- [PN] [AN] [ZNh] [ZNl] [ZS]
				local zoneNumHigh, zoneNumLow, zoneStatus = data:match("^%x%x%x%x(%x%x)(%x%x)(%x%x)$")
				local zoneNum = tonumber(zoneNumHigh .. zoneNumLow, 16)
				debug("(ConcordPlugin::MESSAGE_HANDLERS) zoneNum = "..zoneNum..", zoneStatus = "..zoneStatus)

				-- Add entries in the Zones table.
				if (g_zones[zoneNum] ~= nil) then
					zoneStatus = tonumber(zoneStatus, 16)
					updateZone(zoneNum, zoneStatus)
				else
					task("New zone detected. Reload Luup to create it.", TASK_BUSY_PERM)
				end

				return true
			end,
	["22"] = function (data, subtype)
				if (subtype == "01") then -- Arming Level
					debug("(ConcordPlugin::MESSAGE_HANDLERS) 22/01 = Arming Level")

					-- [PN] [AN] [UNh] [UNl] [AL]
					local partNum, armingLevel = data:match("^(%x%x)%x%x%x%x%x%x(%x%x)$")
					partNum = tonumber(partNum, 16)
					if (g_partitions[partNum] ~= nil) then
						debug("(ConcordPlugin::MESSAGE_HANDLERS) partNum = " .. partNum .. ", armingLevel = " .. armingLevel)
						updatePartition(partNum, armingLevel)
					end
				elseif (subtype == "02") then -- Alarm/Trouble
					debug("(ConcordPlugin::MESSAGE_HANDLERS) 22/02 = Alarm/Trouble")

					-- [PN] [AN] [ST] [SNh] [SNm] [SNl] [GT] [ST] [ESh] [ESl]
					local partNum, srcType, srcNumHigh, srcNumMed, srcNumLow, generalType, specificType = data:match("^(%x%x)%x%x(%x%x)(%x%x)(%x%x)(%x%x)(%x%x)(%x%x)")
					debug("(ConcordPlugin::MESSAGE_HANDLERS) partNum = "..partNum..", srcType = "..srcType..", srcNum = "..srcNumHigh..srcNumMed..srcNumLow..", generalType = "..generalType..", specificType = "..specificType)

					partNum = tonumber(partNum, 16)
					-- Only bus devices and zone alarm sources.
					if (srcType == "00" or srcType == "02") then
						if (generalType == "01") then -- Alarm!
							luup.variable_set(PARTITION_SID, "Alarm", "Active", g_partitions[partNum].id)
							local zoneName = ""
							if (srcType == "02") then
								local zoneNum = tonumber(srcNumHigh .. srcNumMed .. srcNumLow, 16)
								zoneName = luup.devices[g_zones[zoneNum].id].description
							end
							task( "(".. zoneName ..") ALARM: ".. ALARM_TYPES[specificType] .."! ".. os.date(), TASK_ERROR_PERM)
						elseif (generalType == "02") then -- Alarm cleared
							clearTask()
							luup.variable_set(PARTITION_SID, "Alarm", "None", g_partitions[partNum].id)
						end
					end
				elseif (subtype == "0C") then -- Feature State
					debug("(ConcordPlugin::MESSAGE_HANDLERS) 22/0C = Feature State")

					-- [PN] [AN] [FS1]
					local partNum, features = data:match("^(%x%x)%x%x(%x%x)")
					partNum = tonumber(partNum, 16)
					if (g_partitions[partNum] ~= nil) then
						debug("(ConcordPlugin::MESSAGE_HANDLERS) partNum = ".. partNum ..", features = ".. features)
						features = tonumber(features, 16)

						-- Chime mode
						local chimeEnabled = checkBit(features, FEATURE_CHIME) and "1" or "0"
						luup.variable_set(PARTITION_SID, "ChimeEnabled", chimeEnabled, g_partitions[partNum].id)

						-- Quick arm
						QUICK_ARM = checkBit(features, FEATURE_QUICK_ARM)
						local quickArm = QUICK_ARM and "enabled" or "disabled"
						luup.variable_set(PARTITION_SID, "QuickArm", quickArm, g_partitions[partNum].id)
					end
				end
			end
}


-------------------------------------------------------------------------------
-- The 'handleIncoming' function is implemented as a finite-state machine (FSM)
-- Declare here the global variables of the FSM.
-------------------------------------------------------------------------------

local state = 0
local bytesToReceive = 0
local messageLength = ""
local messageType = ""
local messageSubtype = ""
local data = ""
local chksum = ""

function handleIncoming (s)

	if (s == CHAR_ACK) then
		table.remove(messageQueue, 1) -- The last sent message was acknowledged, so remove it from the queue.
		sendAttempt = 0
		return
	elseif (s == CHAR_NAK) then
		table.remove(messageQueue, 1) -- The last sent message was invalid, so remove it from the queue.
		debug("(ConcordPlugin::handleIncoming) Invalid message sent.")
	elseif (s == CHAR_LF) then -- New message
		--debug("(ConcordPlugin::handleIncoming) State #"..state..": new message.")
		state = 1
	elseif (state == 0) then
		return
	elseif (state == 1) then -- Message length first nibble
		--debug("(ConcordPlugin::handleIncoming) State #"..state..": message length, first nibble.")
		messageLength = s
		state = 2
	elseif (state == 2) then -- Message length second nibble
		--debug("(ConcordPlugin::handleIncoming) State #"..state..": message length, second nibble.")
		messageLength = messageLength .. s
		bytesToReceive = tonumber(messageLength, 16)
		state = 3
	elseif (state == 3) then -- Message type first nibble
		--debug("(ConcordPlugin::handleIncoming) State #"..state..": message type, first nibble.")
		messageType = s
		state = 4
	elseif (state == 4) then -- Message type second nibble
		--debug("(ConcordPlugin::handleIncoming) State #"..state..": message type, second nibble.")
		messageType = messageType .. s
		messageSubtype = ""
		data = ""
		bytesToReceive = bytesToReceive - 1
		if (messageType == "22" or messageType == "23") then -- Messages of types 22 and 23 have subtypes.
			state = 5
		else
			state = (bytesToReceive > 1) and 7 or 9 -- If we don't have any data bytes jump to the checkusm.
		end
	elseif (state == 5) then -- Message subtype first nibble
		--debug("(ConcordPlugin::handleIncoming) State #"..state..": data byte, first nibble.")
		messageSubtype = messageSubtype .. s
		state = 6
	elseif (state == 6) then -- Message subtype second nibble
		--debug("(ConcordPlugin::handleIncoming) State #"..state..": data byte, first nibble.")
		messageSubtype = messageSubtype .. s
		bytesToReceive = bytesToReceive - 1
		state = (bytesToReceive > 1) and 7 or 9 -- If we don't have any data bytes jump to the checkusm.
	elseif (state == 7) then -- Data byte first nibble
		--debug("(ConcordPlugin::handleIncoming) State #"..state..": data byte, first nibble.")
		data = data .. s
		state = 8
	elseif (state == 8) then -- Data byte second nibble
		--debug("(ConcordPlugin::handleIncoming) State #"..state..": data byte, second nibble.")
		data = data .. s
		bytesToReceive = bytesToReceive - 1
		state = (bytesToReceive > 1) and 7 or 9 -- If we don't have any more data bytes jump to the checkusm.
	elseif (state == 9) then -- Checksum first nibble
		--debug("(ConcordPlugin::handleIncoming) State #"..state..": checksum, first nibble.")
		chksum = s
		state = 10
	elseif (state == 10) then -- Checksum second nibble
		--debug("(ConcordPlugin::handleIncoming) State #"..state..": checksum, second nibble.")
		if ((messageType == "22" and (messageSubtype == "03" or messageSubtype == "04" or messageSubtype == "05" or messageSubtype == "06" or messageSubtype == "09" or messageSubtype == "0B" or messageSubtype == "0D" or messageSubtype == "0E")) or messageType == "23") then
			-- We're not interested in these messages, so there's no point in verifying the checksum.
			sendChar(CHAR_ACK)
			return
		end

		chksum = chksum .. s

		local message = messageLength .. messageType .. messageSubtype .. data
		debug("(ConcordPlugin::handleIncoming) message = '" .. message .. chksum .. "'.")

		local cchksum = checksum(message)
		if (cchksum == tonumber(chksum, 16)) then
			sendChar(CHAR_ACK)
			local processMessage = MESSAGE_HANDLERS[messageType]
			if (processMessage == nil) then
				debug("(ConcordPlugin::handleIncoming) ERROR: Unhandled message type '" .. messageType .. "'.")
			else
				processMessage(data, messageSubtype)
			end
			state = 0
		else
			sendChar(CHAR_NAK)
			state = 0
		end
	else
		debug("(ConcordPlugin::handleIncoming) State #"..state..": invalid state.")
	end
end


-------------------------------------------------------------------------------
-- Action implementations
-------------------------------------------------------------------------------

function requestArmMode (device, state, pinCode)

	local action = (state == "Disarmed") and "disarm" or "arm"

	if (pinCode == "undefined") then
		pinCode = nil
	end

	if (QUICK_ARM == false or action == "disarm") then
		-- Check if the PIN code exists and if it's valid.
		if (pinCode == nil or pinCode:find("^%d%d%d%d$") == nil) then
			task("Invalid access code. Cannot ".. action .." partition.", TASK_ERROR)
			return
		end
	end

	if (QUICK_ARM == true and action == "arm") then
		pinCode = nil
	end


	local partNum = getPartitionNumber(device)
	if (partNum == nil) then
		task("Failed to get partition number. Cannot ".. action .." partition.", TASK_ERROR)
		return
	end
	debug("(ConcordPlugin::requestArmMode) Request arm mode '".. state .."' for partition #".. partNum ..".")

	local prefix = (state:find("Silent", 1, true) ~= nil) and "05" or ""
	local suffix = (state:find("Instant", 1, true) ~= nil) and "04" or ""

	-- 40h [PN] [AN] [KP1]  [KPn]
	local command = "40" .. padLeft(partNum) .."00".. prefix .. COMMANDS[state] .. suffix
	addToQueue(command)
    command = "40" .. padLeft(partNum) .."00".. expand(pinCode)
    addToQueue(command)

	task("Command to ".. action .." partition #".. partNum .." sent.", TASK_BUSY)
end


function requestQuickArmMode (device, state)
	requestArmMode(device, state)
end


function requestPanicMode (device, state)

	local partNum = getPartitionNumber(device)
	if (partNum == nil) then
		task("Failed to get partition number. Cannot activate panic alarm.", TASK_ERROR)
		return
	end
	debug("(ConcordPlugin::requestPanicMode) Request '".. state .."' panic alarm from partition #".. partNum ..".")
	local command = "40".. padLeft(partNum) .."00".. COMMANDS[state]
	addToQueue(command)

	task( state.." panic alarm command sent.", TASK_BUSY)
end


function toggleChimeMode (device)

	if (EXTRA_FEATURES == false) then
		task("ExtraFeatures is disabled. Toggle Chime Mode command not sent.", TASK_ERROR)
		return
	end

	local partNum = getPartitionNumber(device)
	if (partNum == nil) then
		task("Failed to get partition number. Cannot toggle chime mode.", TASK_ERROR)
		return
	end
	debug("(ConcordPlugin::toggleChimeMode) Toggle chime mode for partition #".. partNum ..".")

	local command = "40".. padLeft(partNum) .."00".. COMMANDS["Chime"]
	addToQueue(command)

	task( "Toggle chime mode command sent.", TASK_BUSY)
end


function bypassZone (device, bypass, pinCode)

	local action = (bypass == "1") and "bypass" or "arm"

	if (pinCode == "undefined") then
		pinCode = nil
	end

	-- Check if the PIN code exists and if it's valid.
	if (pinCode == nil or pinCode:find("^%d%d%d%d$") == nil) then
		task("Invalid access code. Cannot ".. action .." zone.", TASK_ERROR)
		return
	end

	local zoneNum = getZoneNumber(device)
	if (zoneNum == nil) then
		task("Failed to get zone number. Cannot ".. action .." zone.", TASK_ERROR)
		return
	end
	debug("(ConcordPlugin::bypassZone) ".. action .." zone #".. zoneNum ..".")

	local command = "40".. padLeft( g_zones[zoneNum].partNum ) .."00".. COMMANDS["Bypass"]   -- The bypass command
	addToQueue(command)
	command = "40".. padLeft( g_zones[zoneNum].partNum ) .."00".. expand(pinCode)            -- The pin code
	addToQueue(command)
	command = "40".. padLeft( g_zones[zoneNum].partNum ) .."00".. expand( padLeft(zoneNum) ) -- The zone number to be bypassed
	addToQueue(command)

	task( "Command to ".. action .." zone #".. zoneNum .." sent.", TASK_BUSY)
end

-------------------------------------------------------------------------------
-- Init
-------------------------------------------------------------------------------

function init (lul_device)
	log("(ConcordPlugin::init) Concord alarm panel plugin version " .. VERSION .. ".")

	lug_device = lul_device

	g_taskHandle = luup.task("Starting up...", 1, "Concord Plugin", -1)

	-- Check if Debug Mode is enabled.
	local debugMode = luup.variable_get(CONCORD_SID, "DebugMode", lul_device) or ""
	if (debugMode == "") then
		debugMode = DEBUG_MODE and "1" or "0"
		luup.variable_set (CONCORD_SID, "DebugMode", debugMode, lul_device)
	else
		DEBUG_MODE = (debugMode == "1") and true or false
	end

	-- Check if the panel is connected through the serial port, or via Ethernet.
	local attr_ip = luup.devices[lul_device].ip or ""
	if (attr_ip == "") then
		debug ("(ConcordPlugin::init) Running on Serial.")
	else
		local ip, port = attr_ip:match("^([%d.]+):?(%d*)$")
		debug("(ConcordPlugin::init) ip = '"..(ip or "nil").."', port = '"..(port or "nil").."'.")

		-- Validate the IP.
		if (ip:match("^%d+%.%d+%.%d+%.%d+$") ~= nil) then
			if (port == nil or port == "") then
				port = 5000
			end
			debug("(ConcordPlugin::init) Running on Ethernet. IP = "..ip..", port = "..port..".")
			luup.io.open(lul_device, ip, port)
		else
			debug("(ConcordPlugin::init) Invalid IP. Running on Serial.")
		end
	end

	-- Get the number of partitions and add them in the partitions list.
	local partitions = luup.variable_get(CONCORD_SID, "Partitions", lul_device) or ""
	if (partitions == "") then
		luup.variable_set(CONCORD_SID, "Partitions", "1", lul_device)
		g_partitions[1] = {} -- Create partition 1, which it's the default active partition.
	else
		-- Create active partitions (e.g. 1,4).
		for part in partitions:gmatch("%d") do
			g_partitions[tonumber(part, 10)] = {}
		end
	end

	-- Check if Extra Features are enabled.
	-- Currently the only extra feature supported is chime mode.
	local extraFeatures = luup.variable_get(CONCORD_SID, "ExtraFeatures", lul_device) or ""
	if (extraFeatures == "") then
		extraFeatures = EXTRA_FEATURES and "1" or "0"
		luup.variable_set(CONCORD_SID, "ExtraFeatures", extraFeatures, lul_device)
	else
		EXTRA_FEATURES = (extraFeatures == "1") and true or false
	end

	-- Request zone data.
	addToQueue("0203")

	debug("(ConcordPlugin::init) SUCCESS: startup successful.")
	return true, "startup successful", "Concord Plugin"
end
